(**************************************************************************)
(*                                                                        *)
(*     A tutorial on propositional logic and its properties               *)
(*     Copyright (C) 2013                                                 *)
(*                                                                        *)
(*                                                                        *)
(*     Julien Forest                                                      *)
(*                                                                        *)
(*     Cedric-ENSIIE                                                      *)
(*                                                                        *)
(*                                                                        *)
(*   This file is distributed under the terms of the CeCILL-C licence     *)
(*                                                                        *)
(**************************************************************************)

(** This module introduce the propositional formulas and some basic 
    properties on them.
*)

Require Import List.
Require Import Arith.
Require Import Utf8_core. (* Just to have some utf symbols *)

  Reserved Notation "x ⊢ y" (at level 300, no associativity).
  Reserved Notation "⊥" (at level 250, no associativity).
  Reserved Notation "⊧" (at level 300,no associativity).

(** Formulas are defined as an inductive set *)
Inductive Formula : Set := 
  | Bot : Formula     
  | Var : nat -> Formula  (* variable are just integers *)
  | FAnd : Formula -> Formula -> Formula
  | FOr : Formula -> Formula -> Formula
  | FImpl : Formula -> Formula -> Formula
  | FNeg : Formula -> Formula
.

Notation "A → B" := (FImpl A B) : Logic_scope.
Notation  "A ∧ B" := (FAnd A B) : Logic_scope.
Notation  "A ∨ B" := (FOr A B) : Logic_scope.
Notation  "¬ A" := (FNeg A) : Logic_scope.
Notation  "⊥" := Bot : Logic_scope.
Open Scope Logic_scope.

Notation  "⊤" := (FNeg Bot) : Logic_scope.


(* We can construct big ∨ *)
Function build_big_or l := 
  match l with 
    | nil => (⊥)
    | F::l =>  F ∨ (build_big_or l)
  end.

(* And big ∧ *)
Function build_big_and l := 
  match l with 
    | nil => (⊤)
    | F::l =>  F ∧ (build_big_and l)
  end.


(** Equivalence is just define as the conjonction of 
    two implications
*)
Definition FEquiv  F G := FAnd (FImpl F G) (FImpl G F).


(** One can decide equality on formulas *)
Lemma Formula_dec :  forall x y : Formula, {x = y} + {x <> y}.
Proof.
  decide equality.
  decide equality.
Qed.

(** we can then define the set of variables appearing in 
    a given formula by induction over this formula. 

    Since we only want to be able to test if a variable belongs to 
    this set and we don't care about efficiency, we can use list to 
    modelize sets
*)
Function var_set F := 
  match F with 
      | Bot => nil
      | Var n => n::nil
      | FAnd F1 F2 | FOr F1 F2 | FImpl F1 F2 => 
                                 (var_set F1)++(var_set F2)
      | FNeg F => var_set F 
  end.
  

(** Another point of view is to define the propertie for a variable to 
    appear in a formula
*)
Inductive appears_in_formula (x:nat) : Formula -> Prop :=
| Formula_var : appears_in_formula x (Var x)
| Formula_and_l : forall F G, appears_in_formula x F -> 
                              appears_in_formula x (FAnd F G)
| Formula_and_r : forall F G, appears_in_formula x G -> 
                              appears_in_formula x (FAnd F G)
| Formula_or_l : forall F G, appears_in_formula x F -> 
                             appears_in_formula x (FOr F G)
| Formula_or_r : forall F G, appears_in_formula x G -> 
                             appears_in_formula x (FOr F G)
| Formula_impl_l : forall F G, appears_in_formula x F -> 
                               appears_in_formula x (FImpl F G)
| Formula_impl_r : forall F G, appears_in_formula x G -> 
                               appears_in_formula x (FImpl F G)
| Formula_neg : forall F, appears_in_formula x F -> 
                          appears_in_formula x (FNeg F)
.

(* Those two notions are equivalent *)
Lemma var_set_equiv_appears_in_formula : 
  forall F x, 
    (In x (var_set F)) <->
    (appears_in_formula x F).
Proof.
  (* the proof is done by induction on the definition of 
     var_set thanks to Function *)
  intros F.
  functional induction (var_set F) 
    as [ |  _F n IHn   (* we name generated hypothesis *)
         | _F F1 F2 _F' IHF1 IHF2
         | _F F1 F2 _F' IHF1 IHF2 | _F F1 F2 _F' IHF1 IHF2 |
         _F' F _F'' IHF ];intro x.

  - (* Top case *)
    split;intro abs;inversion abs.
 
  - (* Var case *)
    split. (* we split equivalence into two implications *)
    + intro H;simpl in H;destruct H as [H|H].
      * subst;apply Formula_var.
      * tauto.
    + intros H;inversion_clear H;simpl;auto.

  - (* And case *)
    split;intro H.
    + apply in_app_or in H.
      destruct H as [H|H].
      * apply Formula_and_l.
        rewrite IHF1 in H. 
        assumption.
      * apply Formula_and_r.
        rewrite IHF2 in H. 
        assumption.
    + inversion_clear H.
      * apply in_or_app.
        left.
        rewrite IHF1.
        assumption.
      * apply in_or_app.
        right.
        rewrite IHF2.
        assumption.

  - (* Or case *)
    split;intro H.
    + apply in_app_or in H.
      destruct H as [H|H].
      * apply Formula_or_l.
        rewrite IHF1 in H. 
        assumption.
      * apply Formula_or_r.
        rewrite IHF2 in H. 
        assumption.
    + inversion_clear H.
      * apply in_or_app.
        left.
        rewrite IHF1.
        assumption.
      * apply in_or_app.
        right.
        rewrite IHF2.
        assumption.

  - (* Impl case *)
    split;intro H.
    + apply in_app_or in H.
      destruct H as [H|H].
      * apply Formula_impl_l.
        rewrite IHF1 in H. 
        assumption.
      * apply Formula_impl_r.
        rewrite IHF2 in H. 
        assumption.
    + inversion_clear H.
      * apply in_or_app.
        left.
        rewrite IHF1.
        assumption.
      * apply in_or_app.
        right.
        rewrite IHF2.
        assumption.

  - (* Neg case *)
    split;intro H.
    + rewrite IHF in H.
      apply Formula_neg.
      assumption.
    + inversion_clear H.
      rewrite IHF.
      assumption.
Qed.

(** We then define the replacement of a variable [p] by a formula [G] in a 
    formula [F]
*)
Function freplace (F:Formula) (p:nat) (G:Formula) {struct F} : Formula :=
  match F with 
      | Bot => Bot
      | Var q => if eq_nat_dec p q then G else F
      | FAnd F1 F2 => FAnd (freplace F1 p G) (freplace F2 p G)
      | FOr F1 F2 => FOr (freplace F1 p G) (freplace F2 p G)
      | FImpl F1 F2 => FImpl (freplace F1 p G) (freplace F2 p G)
      | FNeg F => FNeg (freplace F p G)
  end.

(* We then define the property for a formula [F] to be a sub formula of 
   another formula [G]
*)
Inductive sub_formula (F:Formula) : Formula -> Prop :=
  | Sub_formula_refl : sub_formula F F
  | sub_formula_and_l : forall G H, sub_formula F G -> 
                                    sub_formula F (FAnd G H)
  | sub_formula_or_l : forall G H, sub_formula F G -> 
                                   sub_formula F (FOr G H)
  | sub_formula_impl_l : forall G H, sub_formula F G -> 
                                     sub_formula F (FImpl G H)
  | sub_formula_neg : forall G, sub_formula F G -> 
                                sub_formula F (FNeg G)
.
