(**************************************************************************)
(*                                                                        *)
(*     A tutorial on propositional logic and its properties               *)
(*     Copyright (C) 2013                                                 *)
(*                                                                        *)
(*                                                                        *)
(*     Julien Forest                                                      *)
(*                                                                        *)
(*     Cedric-ENSIIE                                                      *)
(*                                                                        *)
(*                                                                        *)
(*   This file is distributed under the terms of the CeCILL-C licence     *)
(*                                                                        *)
(**************************************************************************)

(** 
    This module define the natural deduction of propositional logic 
    and proves that it is correct and complete w.r.t. semantic
*)
Require Import Formula.
Require Import Semantic.
Require Import List.
Require Import Arith.
Require Import Omega.


Inductive natural_deduction_proof : list Formula -> Formula -> Prop := 
  | Ax : forall Γ A, In A Γ -> natural_deduction_proof Γ A
  | Aff : (* we use the generalized version of Aff *)
      forall Γ Σ A, (incl Γ Σ) -> natural_deduction_proof Γ A -> natural_deduction_proof Σ A
  | Impl_i : 
      forall Γ A B, natural_deduction_proof (A::Γ) B -> natural_deduction_proof Γ  (A → B)
  | Impl_e : forall Γ A B, 
               natural_deduction_proof Γ (A → B) -> 
               natural_deduction_proof Γ A -> 
               natural_deduction_proof Γ B
  | And_i : forall Γ A B, natural_deduction_proof Γ A -> natural_deduction_proof Γ B -> 
                          natural_deduction_proof Γ (A ∧ B)
  | And_e_l : forall Γ A B, natural_deduction_proof Γ (A ∧ B) -> natural_deduction_proof Γ A
  | And_e_r : forall Γ A B, natural_deduction_proof Γ (A ∧ B) -> natural_deduction_proof Γ B
  | Or_i_l : forall Γ A B, natural_deduction_proof Γ A -> natural_deduction_proof Γ (A ∨ B)
  | Or_i_r : forall Γ A B, natural_deduction_proof Γ B -> natural_deduction_proof Γ (A ∨ B)
  | Or_e : forall Γ A B C, natural_deduction_proof Γ (A ∨ B) -> 
                               natural_deduction_proof (A::Γ) C -> 
                               natural_deduction_proof (B::Γ) C -> 
                               natural_deduction_proof Γ C
  | Neg_i : forall Γ A, natural_deduction_proof (A::Γ) (⊥) -> 
                            natural_deduction_proof Γ (¬ A)
  | Neg_e : forall Γ A, natural_deduction_proof Γ (¬ A) -> 
                            natural_deduction_proof Γ A -> 
                            natural_deduction_proof Γ (⊥)
  | Abs_c : forall Γ A, natural_deduction_proof ((¬ A) :: Γ) (⊥) -> 
                            natural_deduction_proof Γ A
.

Notation " A ⊢ B " := (natural_deduction_proof A B) : Logic_scope.

(* We can prove some classical rules from 
   our system
*)      

Lemma Proof_true Γ : (Γ ⊢ ⊤).
Proof.
  apply Neg_i.
  apply Ax;simpl;auto.
Qed.

Lemma A_or_not_A : forall Γ A, (Γ ⊢ (A ∨ ¬ A)).
Proof.
  intros Γ A.
  apply Abs_c.
  apply Neg_e with A.

  - apply Neg_i.
    apply Neg_e with (A ∨ ¬ A).
    * apply Ax;simpl;auto.
    * apply Or_i_l.
      apply Ax;simpl;auto.

  - apply Abs_c.
    apply Neg_e with (A ∨ ¬ A).
    * apply Ax;simpl;auto.
    * apply Or_i_r.
      apply Ax;simpl;auto.
Qed.

Lemma Pierce Γ A :  (((¬ A)::Γ) ⊢ A) -> (Γ ⊢ A).
Proof.
  intros H.
  apply Or_e with A (¬ A).
  - apply A_or_not_A.
  - apply Ax;simpl;auto.
  - assumption.
Qed.

Lemma neg_g Γ A :  In A Γ -> In (¬A) Γ -> (Γ ⊢ ⊥).
Proof.
  intros H H1.
  apply Neg_e with A.
  - apply Ax;simpl;auto.
  - apply Ax;simpl;auto.
Qed.


Lemma contradiction Γ A :
  (Γ ⊢ ⊥) -> (Γ ⊢ A).
Proof.
  intros H.
  apply Abs_c.
  apply Aff with Γ.
  - red. simpl. auto.
  - exact H.
Qed.

Lemma contraposition  Γ A B : 
  (Γ ⊢ (¬B → ¬A )) ->  (Γ ⊢ (A → B)).
Proof.
  intros H.
  apply Impl_i.
  apply Abs_c.
  apply Neg_e with A.
  - apply Impl_e with (¬ B).
    + apply Aff with Γ. 
      * intros F;simpl;auto.
      * assumption.
    + apply Ax;simpl;auto.
  - apply Ax;simpl;auto.
Qed.

Lemma excluded_middle Γ A B : 
  ((A::Γ) ⊢ B) -> ((FNeg A :: Γ)⊢ B) -> (Γ ⊢ B).
Proof.
  intros H H0.
  apply Or_e with A (¬ A).
  - apply A_or_not_A.
  - assumption.
  - assumption.
Qed.


(* We can even extend Or_e to big ∨ *)
Lemma Extended_Or_e Γ F l : 
   (Γ ⊢ build_big_or l) ->
   (forall F', In F' l -> natural_deduction_proof (F'::Γ) F) ->
   (Γ ⊢ F).
Proof.
  revert Γ.
  induction l as [ | G l IHl].

  - simpl.
    intros Γ H H0.
    apply contradiction.
    assumption.

  - simpl; intros Γ H H1.
    apply Or_e with G (build_big_or l).
    + assumption.
    + auto.
    + apply IHl.
      * apply Ax;simpl;auto.
      * intros F' H2.
        apply Impl_e with F'.
        { apply Aff with Γ;[auto with *|].
          apply Impl_i.
          apply H1;auto.
        }          
        { apply Ax;simpl;auto. }
Qed.

(* And some associativity rules *)
Lemma Proof_Assoc_or Γ F₁ F₂ F₃:
  ( Γ ⊢  F₁ ∨ (F₂ ∨ F₃)) <-> 
  ( Γ ⊢ (F₁ ∨ F₂) ∨ F₃).
Proof.                         
  split;intros .

  - apply Or_e with F₁ (F₂∨F₃).
    + assumption.
    + apply Or_i_l;apply Or_i_l;apply Ax;simpl;auto.
    + apply Or_e with F₂ F₃.
      * apply Ax;simpl;auto.
      * apply Or_i_l;apply Or_i_r;apply Ax;simpl;auto.
      * apply Or_i_r;apply Ax;simpl;auto.

  - apply Or_e with (F₁ ∨ F₂) F₃.
    + assumption.
    + apply Or_e with F₁ F₂.
      * apply Ax;simpl;auto.
      * apply Or_i_l;apply Ax;simpl;auto.
      * apply Or_i_r;apply Or_i_l;apply Ax;simpl;auto.
    + apply Or_i_r;apply Or_i_r;apply Ax;simpl;auto.
Qed.

Lemma Proof_Assoc_and Γ F₁ F₂ F₃:
  ( Γ ⊢  F₁ ∧ (F₂ ∧ F₃)) <-> 
  ( Γ ⊢ (F₁ ∧ F₂) ∧ F₃).
Proof.                         
  split;intros .

  - apply And_i.
    + apply And_i.
      * apply And_e_l with (F₂∧F₃).
        assumption.
      * apply And_e_l with F₃.
        apply And_e_r with F₁.
        assumption.
    + apply And_e_r with F₂.
      apply And_e_r with F₁.
      assumption.

  - apply And_i.
    + apply And_e_l with F₂.
      apply And_e_l with F₃.
      assumption.
    + apply And_i.
      * apply And_e_r with F₁.
        apply And_e_l with F₃.
        assumption.
      * apply And_e_r with (F₁∧F₂).
        assumption.
Qed.


Lemma Proof_build_big_or_app : 
  forall Γ l1 l2 , 
     (Γ ⊢ build_big_or (l1++l2)) <-> 
     (Γ ⊢ (build_big_or l1) ∨ (build_big_or l2)).
Proof.
  intros Γ l1;revert Γ.
  induction l1 as [ | F l1 IH];intros Γ l2.
  
  - simpl.
    split;intro H.
    + apply Or_i_r;assumption.
    + apply Or_e with (⊥) (build_big_or l2).
      * assumption.
      * apply contradiction. apply Ax;simpl;auto.
      * apply Ax;simpl;auto.

  - simpl.
    rewrite <- Proof_Assoc_or.
    split;intros H.
    + apply Or_e with F (build_big_or (l1 ++ l2)).
      * assumption.
      * apply Or_i_l;apply Ax;simpl;auto. 
      * apply Or_i_r.
          rewrite <- IH.
          apply Ax;simpl;auto.

    + apply Or_e with F (build_big_or l1 ∨ build_big_or l2).
      * assumption.
      * apply Or_i_l;apply Ax;simpl;auto.
      * apply Or_i_r. rewrite IH.
        apply Ax;simpl;auto.
Qed.


Lemma Proof_Or_and_distrib Γ F₁ F₂ F₃ : 
  (Γ ⊢ F₁ ∧ F₂ ∨ F₁ ∧ F₃) <-> 
  (Γ ⊢ F₁ ∧ (F₂ ∨  F₃)).
Proof.
  split.

  - intros H.
    apply Or_e with (F₁ ∧ F₂) (F₁ ∧ F₃).
    + assumption.
    + apply And_i. 
      * apply And_e_l with F₂;apply Ax;simpl;auto.
      * apply Or_i_l;apply And_e_r with F₁;apply Ax;simpl;auto.
    + apply And_i. 
      * apply And_e_l with F₃;apply Ax;simpl;auto.
      * apply Or_i_r;apply And_e_r with F₁;apply Ax;simpl;auto.

  - intros H0.
    apply Or_e with F₂ F₃.
    + apply And_e_r with F₁.
      assumption.
    + apply Or_i_l.
      apply And_i.
      * apply Aff with Γ;[auto with *|].
        apply And_e_l with (F₂ ∨ F₃).
        assumption.
      * apply Ax;simpl;auto.
    + apply Or_i_r.
      apply And_i.
      * apply Aff with Γ;[auto with *|].
        apply And_e_l with (F₂ ∨ F₃).
        assumption.
      * apply Ax;simpl;auto.
Qed.

Lemma Proof_Or_andl_distrib Γ F l : 
  ( Γ ⊢ build_big_or (List.map (fun b => F ∧ b) l)) <-> 
  ( Γ ⊢ F ∧ (build_big_or l)).
Proof.
  revert Γ.
  induction l as [| G l IH].

  - simpl.
    split.
    + intros H.
      apply contradiction.
      assumption.
    + intros H.
      apply And_e_r with F;assumption.

  - simpl.
    split.
    + intros H.
      apply Or_e with (F ∧ G) (build_big_or (List.map (fun b => FAnd F b) l)).
      * assumption.
      * apply And_i.
        { apply And_e_l with G;apply Ax;simpl;auto. }
        {
          apply Or_i_l.
          apply And_e_r with F;apply Ax;simpl;auto.
        }
      * apply And_i.
        {
          apply And_e_l with (build_big_or l).
          rewrite <- IH.
          apply Ax;simpl;auto.
        }
        {
          apply Or_i_r.
          apply And_e_r with F.
          rewrite <- IH.
          apply Ax;simpl;auto.
        }
    + intros H.
      apply Or_e with G (build_big_or l).
      * apply And_e_r with F.
        assumption.
      * apply Or_i_l.
        apply And_i.
        {
          apply Aff with Γ;[auto with *|].
          apply And_e_l with (G ∨ build_big_or l).
          assumption.
        }
        { apply Ax;simpl;auto. }
      * apply Or_i_r.
        rewrite IH.
        apply And_i.
        {
          apply Aff with Γ;[auto with *|].
          apply And_e_l with (G∨ (build_big_or l)).
          assumption.
        }
        { apply Ax;simpl;auto. }
Qed.

(* Natural deduction correctness proof *)
Lemma natural_deduction_correctness : forall Γ A, (Γ ⊢ A) -> deduce Γ A.
Proof.

  intros Γ A H.
  induction H as [ Γ A H |  (* Ax *)
                   Γ Σ B Hincl H1 IH1 | (* Aff *)
                   Γ A B H1 IH1 |  (* Impl_i *)
                   Γ A B H1 IH1 H2 IH2 |  (* Impl_e *)
                   Γ A B H1 IH1 H2 IH2 |  (* And_i *)
                   Γ A B H1 IH1 | (* And_e_l *)
                   Γ A B H1 IH1 | (* And_e_r *)
                   Γ A B H1 IH1 | (* Or_i_l *)
                   Γ A B H1 IH1 | (* Or_i_r *)
                   Γ A B C H1 IH1 H2 IH2 H3 IH3 | (* Or_e *)
                   Γ A H1 IH1 | (* Neg_i *)
                   Γ A H1 IH1 H2 IH2 | (* Neg_e *)
                   Γ A H1 IH1 ] (* Abs_c *).
  
  - intros I H_I.
    apply H_I; assumption.

  - intros I H_I.
    apply IH1.
    intros F H_F.
    apply H_I.
    apply Hincl.
    auto.

  -  rewrite deduce_impl_intro;auto.

  - intros I H_I.
    apply deduce_impl with A.
    red;simpl.
    intros F [H3 |[H3|H3]].
    + subst;apply IH2;assumption.
    + subst;apply IH1;assumption.
    + tauto.

  - intros I H_I.
    red;simpl.
    assert (HA:=IH1 _ H_I).
    assert (HB:=IH2 _ H_I).
    rewrite HA;rewrite HB;reflexivity.

  - intros I H_I.
    assert (HAB:=IH1 _ H_I).
    red in HAB;simpl in HAB.
    rewrite Bool.andb_true_iff in HAB;destruct HAB;assumption.

  - intros I H_I.
    assert (HAB:=IH1 _ H_I).
    red in HAB;simpl in HAB.
    rewrite Bool.andb_true_iff in HAB;destruct HAB;assumption.

  - intros I H_I.
    red;simpl.
    rewrite (IH1 _ H_I).
    reflexivity.

  - intros I H_I.
    red;simpl.
    rewrite (IH1 _ H_I).
    auto with bool. 

  - intros I H_I.
    assert (H:=IH1 _ H_I).
    red in H;simpl in H.
    destruct (Bool.orb_true_elim _ _ H) as [h1|h1].
    + apply (IH2 I).
      red;simpl.
      intros F [h2|h2].
      subst. exact h1.
      apply H_I.
      assumption.
    + apply (IH3 I).
      red;simpl.
      intros F [h2|h2].
      subst. exact h1.
      apply H_I.
      assumption.

  - intros I H_I.
    red;simpl.
    rewrite deduce_bot_contradictory in IH1.
    destruct (IH1 I) as [F [H2 H3]].
    simpl in H2;destruct H2 as [H2|H2].
    + subst.
      apply Bool.eq_true_not_negb;apply H3.
    + elim H3.
      apply H_I;assumption.

  - intros I H_I.
    absurd (interp I A = true).
    + apply  IH1 in H_I.
      red in H_I;simpl in H_I.
      rewrite Bool.negb_true_iff in H_I.
      intro abs. rewrite H_I in abs;discriminate.

    +  apply IH2 in H_I. apply H_I.

  - rewrite deduce_is_contradictory_neg.
    apply deduce_bot_contradictory;assumption.
Qed.

(* proving F → G is equivalent to prove ¬ F∨ G*)
Lemma natural_deduction_impl_equiv_neg_or : 
  forall Γ F G,  (Γ ⊢ F → G) <->  (Γ ⊢ ¬ F ∨ G).
Proof.
  intros Γ F G.
  split;intro H.

  - apply excluded_middle with F.
    + apply Or_i_r.
      apply Impl_e with F .
      * apply Aff with Γ. 
        intro Φ;simpl;auto.
        auto.
      * apply Ax;simpl;auto.
    + apply Or_i_l;apply Ax;simpl;auto.

  - apply Impl_i.
    apply Or_e with (FNeg F) G.
    + apply Aff with Γ.
      * red;simpl;auto.
      * assumption.
    + apply Abs_c.
      apply neg_g with F.
      * simpl;auto.
      * simpl;auto.
    + apply Ax;simpl;auto.
Qed.

(* Proving ¬(F∧G) is equivalent to prove ¬ F ∨ ¬ G *)
Lemma natural_deduction_neg_and_equiv_or_neg Γ F G : 
 (Γ ⊢ ¬(F∧G)) <-> ( Γ ⊢ ¬ F ∨ ¬ G).
Proof.
  split;intro H.

  - apply excluded_middle with F.
    + apply Or_i_r.
      apply Neg_i.
      apply Neg_e with (F ∧ G).
      * apply Aff with Γ.
        red;simpl;auto.
        assumption.
      * apply And_i;apply Ax;simpl;auto.
    + apply Or_i_l;apply Ax;simpl;auto.

  - apply Neg_i.
    apply Or_e with (¬ F) (¬ G).
    + apply Aff with Γ. 
      * red;simpl;auto.
      * assumption.
    + apply Neg_e with F.
      * apply Ax;simpl;auto.
      * apply And_e_l with G;apply Ax;simpl;auto.
    + apply Neg_e with G.
      * apply Ax;simpl;auto.
      * apply And_e_r with F;apply Ax;simpl;auto.
Qed.

(* Proving ¬(F∨G) is equivalent to prove ¬ F ∧ ¬ G *)
Lemma natural_deduction_neg_or_equiv_and_neg Γ F G : 
  (Γ ⊢ ¬ (F ∨ G)) <-> (Γ ⊢ ¬ F ∧ ¬ G).
Proof.
  split;intro H.

  - apply And_i.
    +  apply Neg_i.
       apply Neg_e with (F ∨ G).
       * apply Aff with Γ;[red;simpl;auto|assumption].
       * apply Or_i_l;apply Ax;simpl;auto.
    + apply Neg_i.
      apply Neg_e with (F ∨ G).
      * apply Aff with Γ;[red;simpl;auto|assumption]. 
      * apply Or_i_r;apply Ax;simpl;auto.

  - apply Neg_i.
    apply Or_e with F G.
    * apply Ax;simpl;auto.
    * apply Neg_e with F.
      { apply Aff with Γ;[red;simpl;auto|].
        apply And_e_l with (¬ G).
        assumption.
      }        
      { apply Ax;simpl;auto. }     
    * apply Neg_e with G.
      { apply Aff with Γ;[red;simpl;auto|].
        apply And_e_r with (¬ F).
        assumption.
      }        
      { apply Ax;simpl;auto. }     
Qed.

Lemma natural_deduction_or_and_distrib  Γ F G H:
  (Γ ⊢ F ∨ G∧H) <-> (Γ ⊢  (F∨G) ∧ (F ∨ H)).
Proof.
  split;intro h.

  - apply Or_e with F (G ∧ H).
    + assumption.
    + apply And_i;apply Or_i_l;apply Ax;simpl;auto.
    + apply And_i.
      * apply Or_i_r.
        apply And_e_l with H.
        apply Ax;simpl;auto.
      * apply Or_i_r.
        apply And_e_r with G.
        apply Ax;simpl;auto.

  - apply excluded_middle with F.
    + apply Or_i_l;apply Ax;simpl;auto.
    + apply Or_i_r.
      apply And_i.
      * apply Or_e  with F G.
        { apply Aff with Γ;[red;simpl;auto|].
          apply And_e_l with (F ∨ H);assumption.
        }
        {
          apply contradiction.
          apply neg_g with F.
          simpl;auto.
          simpl;auto.
        }
        { apply Ax;simpl;auto.  
        }
      * apply Or_e with F H.
        { apply Aff with Γ;[red;simpl;auto|].
          apply And_e_r with (F ∨ G);assumption.
        }
        {
          apply contradiction.
          apply neg_g with F.
          simpl;auto.
          simpl;auto.
        }
        { apply Ax;simpl;auto.  
        }
Qed.


(* ∨ is commutative in proofs *)
Lemma natural_deduction_or_commutes  Γ A B: 
  (Γ ⊢  A ∨ B) <->   (Γ ⊢ B ∨ A).
Proof.
  split;intro H.

  + apply Or_e with A B.
    * assumption.
    * apply Or_i_r;apply Ax;simpl;auto.
    * apply Or_i_l;apply Ax;simpl;auto.

  + apply Or_e with B A.
    * assumption.
    * apply Or_i_r;apply Ax;simpl;auto.
    * apply Or_i_l;apply Ax;simpl;auto.
Qed.

(**************************************************************************)
(*                          COMPLETENESS PROOF                            *)
(*  We first define the notion of formula built from a non empty word     *) 
(*  over bool as :                                                        *)
(*    Φʷ = pʷ₁ ∧ … ∧ pʷ_n  wher pʷ_i = Var n     if w_i = true and        *)
(*                              pʷ_i = ¬ (Var n) if w_i = false           *)
(*                                                                        *)
(* The proof is then splitted in three parts                              *)
(* 1- We first show that :  ⊢ ⋁_w Φʷ                                      *)
(* 2- We then show that :                                                 *)
(*    a- if Φʷ ⊧ F then Φʷ ⊢ F                                            *)
(*    b- if Φʷ ⊧ ¬F then Φʷ ⊢ ¬F                                          *)
(* 3- We obviouly conclude the proof for tautologies                      *)
(*                                                                        *)
(* Rq: words are modelize as boolean lists                                *)
(**************************************************************************)


(* [build_from_word_from n l] builds Φˡ but with first var starting at n 
   and not 0
*)
Function build_from_word_from n l := 
  match l with 
      | nil => (⊤) 
      | true::l =>  (Var n) ∧ (build_from_word_from (S n) l)
      | false::l =>  ¬ (Var n) ∧ (build_from_word_from (S n) l)
  end.

Lemma build_from_word_from_vars n l : 
  forall m, appears_in_formula m (build_from_word_from n l) <-> 
            m>=n /\ m < n + length l.
Proof.
  functional induction (build_from_word_from n l);simpl;intros m.

  - split. 
    + intros abs;inversion abs.
      inversion H0.
    + intros H.
      apply False_ind;omega.

  - split.
    + intros H.
      inversion_clear H.
      * inversion_clear H0.
        omega.
      * rewrite IHf in H0.
        omega.

    + intros H.
      case (eq_nat_dec n m).
      * intros H1.
        subst.
        apply Formula_and_l.
        apply Formula_var.
      * intros H1.
        apply Formula_and_r.
        rewrite IHf.
        omega.

  - split.
    + intros H.
      inversion_clear H.
      * inversion_clear H0.
        inversion_clear H.
        omega.
      * rewrite IHf in H0.
        omega.

    + intros H.
      case (eq_nat_dec n m).
      * intros H1.
        subst.
        apply Formula_and_l.
        apply Formula_neg.
        apply Formula_var.
      * intros H1.
        apply Formula_and_r.
        rewrite IHf.
        omega.
Qed.

(* build_from_word is just build_from_word_from 0 *)
Definition build_from_word := build_from_word_from 0.


(* we will need the following technical lemma on inclusion 
*)
Lemma incl_destruct_app A (l m n:list A) : incl (l++m) n -> incl l n /\ incl m n.
Proof.
  revert m n.
  induction l as [|x l].

  - simpl.
    intros m n H.
    split.
    + red. simpl;tauto.
    + assumption.

  - intros m n H.
    split.
    + red. simpl.
      intros a [H1| H1].
      * subst.
        apply H.
        simpl;auto.
      *  apply H;simpl;auto with *.
    +  intros a H1.
       apply H.
       apply in_or_app;auto.
Qed.


(* Giwen a word w, we can define an interpretation 
   such that I(p_i)=w_i. if i<length w and I(p) = true if p >= length w

   We first define the interpretation starting at variable n*)
Fixpoint interp_from_word_from n w := 
  match w with 
      | nil => (fun m => true)
      | x::w => (fun m => if Peano_dec.eq_nat_dec n m 
                          then x
                          else interp_from_word_from (S n) w m
                )
  end.

(* The interpretation always return true outside the word w *)
Lemma interp_from_word_from_not_in_word : 
  forall w n m, n < m \/  n>=n+length w  -> 
                interp_from_word_from m w n = true.
Proof.
  induction w as [| x w IH].

  - simpl;reflexivity.

  - simpl.
    intros n m H.
    case (eq_nat_dec m n).  
    + intros H'. 
      apply False_ind. 
      omega.
    + intros H1.
      apply IH.
      omega.
Qed.

(* Of course Iʷ⊧Φʷ *)
Lemma interp_from_word_from_satisfies_build_from_word l : 
  forall n,
    satisfy_formula (interp_from_word_from n l) (build_from_word_from n l).
Proof.
  intro n.
  functional induction (build_from_word_from n l).

  - reflexivity.

  - red;simpl.
    case (eq_nat_dec n n).
    + intros _;simpl. 
      rewrite change_interp with (I':=interp_from_word_from (S n) l0). 
      * apply IHf.
      * intros x H.
        apply build_from_word_from_vars in H.
        case (eq_nat_dec n x).
        { intro H1;apply False_ind;omega. }
        { reflexivity. }
    + intros abs;elim abs;reflexivity.


  - red;simpl.
    case (eq_nat_dec n n).
    + intros _;simpl. 
      rewrite change_interp with (I':=interp_from_word_from (S n) l0). 
      * apply IHf.
      * intros x H.
        apply build_from_word_from_vars in H.
        case (eq_nat_dec n x).
        { intro H1;apply False_ind;omega. }
        { reflexivity. }
    + intros abs;elim abs;reflexivity.
Qed.


(* any variable deducible from Φʷ is provable from Φʷ *)
Lemma var_deduce_build_from_var_proved_build_from m n w:
  deduce (build_from_word_from m w :: nil) (Var n) ->
  (build_from_word_from m w :: nil ⊢ Var n).
Proof.
  functional induction (build_from_word_from m w).

  - intros abs. (* No variable is a tautology *)
    apply False_ind.
    absurd (interp (fun n => false) (Var n) = true).
    + simpl.
      discriminate.
    + apply abs.
      red;simpl.
      intros F [H|H].
      * subst. reflexivity.
      * tauto.

  - intro Hdeduce.
    case (eq_nat_dec n0 n).
    + intro;subst.
      apply And_e_l with (build_from_word_from (S n) l0).
      apply Ax;simpl;auto.
    + intro H1.
      apply Impl_e with (build_from_word_from (S n0) l0).
      * apply Aff with nil;[red;simpl;auto|].
        apply Impl_i.
        apply IHf.
        intros I H_I.
        red.
        rewrite change_interp with (I':=fun x => if eq_nat_dec x n0 then true else I x).       
        { apply Hdeduce.
          red;simpl;intros F [h|h];[subst|tauto].
          red;simpl.
          case (eq_nat_dec n0 n0);[|intro abs;elim abs;reflexivity].
          intros _;simpl.
          rewrite change_interp with (I':=I).
          { apply H_I.
            simpl;auto.
          }
          { intros n1 H0.
            apply build_from_word_from_vars in H0.
            destruct H0.
            case (eq_nat_dec n1 n0).
            intro;apply False_ind;omega.
            reflexivity.
          }
        }
        {
          intros n1 H;inversion_clear H.
          case (eq_nat_dec n n0).
          intros abs;elim H1;subst;reflexivity.
          reflexivity.
        }
      *  apply And_e_r with (Var n0).
         apply Ax;simpl;auto.
       
  - intro Hdeduce.
    case (eq_nat_dec n0 n).
    + intro;subst.
      absurd (satisfy_formula (fun x => if eq_nat_dec x n then false else interp_from_word_from (S n) l0 x) (Var n)).
      * unfold satisfy_formula;simpl.
        case (eq_nat_dec n n);[discriminate|intro abs;elim abs;reflexivity].
      * apply Hdeduce.
        red;simpl;intros F [h|h];[|tauto].
        red;simpl.
        subst;simpl.
        case (eq_nat_dec n n);[intros _;simpl|intro abs;elim abs;reflexivity].
        rewrite change_interp with (I':=interp_from_word_from (S n) l0).
        { apply interp_from_word_from_satisfies_build_from_word. }
        { intros x H.
          apply build_from_word_from_vars in H.
          case (eq_nat_dec x n).
          { intro;apply False_ind;omega. }
          { reflexivity. }
        }
    + intro H1.
      apply Impl_e with (build_from_word_from (S n0) l0).
      * apply Aff with nil;[red;simpl;auto|].
        apply Impl_i.
        apply IHf.
        intros I H_I.
        red.
        rewrite change_interp with (I':=fun x => if eq_nat_dec x n0 then false else I x).       
        { apply Hdeduce.
          red;simpl;intros F [h|h];[subst|tauto].
          red;simpl.
          case (eq_nat_dec n0 n0);[|intro abs;elim abs;reflexivity].
          intros _;simpl.
          rewrite change_interp with (I':=I).
          { apply H_I.
            simpl;auto.
          }
          { intros n1 H0.
            apply build_from_word_from_vars in H0.
            destruct H0.
            case (eq_nat_dec n1 n0).
            intro;apply False_ind;omega.
            reflexivity.
          }
        }
        {
          simpl.
          intros n1 H. 
          inversion_clear H. 
          case (eq_nat_dec n n0).
          intros abs;elim H1;subst; reflexivity.
          reflexivity.
        }
      *  apply And_e_r with (FNeg (Var n0));apply Ax;simpl;auto.
Qed.

(* The same for negative variable *)
Lemma neg_var_deduce_build_from_neg_var_proved_build_from m n w:
  deduce (build_from_word_from m w :: nil) (¬ (Var n)) ->
  (build_from_word_from m w :: nil ⊢ ¬ Var n).
Proof.
  functional induction (build_from_word_from m w).

  - intros abs. (* No variable is a tautology *)
    apply False_ind.
    absurd (interp (fun n => true) (¬ (Var n)) = true).
    + simpl.
      discriminate.
    + apply abs.
      red;simpl.
      intros F [H|H].
      * subst. reflexivity.
      * tauto.

 - intro Hdeduce.
    case (eq_nat_dec n0 n).
    + intro;subst.
      absurd (satisfy_formula (fun x => if eq_nat_dec x n then true else interp_from_word_from (S n) l0 x) (¬ (Var n))).
      * unfold satisfy_formula;simpl.
        case (eq_nat_dec n n);[discriminate|intro abs;elim abs;reflexivity].
      * apply Hdeduce.
        red;simpl;intros F [h|h];[|tauto].
        red;simpl.
        subst;simpl.
        case (eq_nat_dec n n);[intros _;simpl|intro abs;elim abs;reflexivity].
        rewrite change_interp with (I':=interp_from_word_from (S n) l0).
        { apply interp_from_word_from_satisfies_build_from_word. }
        { intros x H.
          apply build_from_word_from_vars in H.
          case (eq_nat_dec x n).
          { intro;apply False_ind;omega. }
          { reflexivity. }
        }
    + intro H1.
      apply Impl_e with (build_from_word_from (S n0) l0).
      * apply Aff with nil;[red;simpl;auto|].
        apply Impl_i.
        apply IHf.
        intros I H_I.
        red.
        rewrite change_interp with (I':=fun x => if eq_nat_dec x n0 then true else I x).       
        { apply Hdeduce.
          red;simpl;intros F [h|h];[subst|tauto].
          red;simpl.
          case (eq_nat_dec n0 n0);[|intro abs;elim abs;reflexivity].
          intros _;simpl.
          rewrite change_interp with (I':=I).
          { apply H_I.
            simpl;auto.
          }
          { intros n1 H0.
            apply build_from_word_from_vars in H0.
            destruct H0.
            case (eq_nat_dec n1 n0).
            intro;apply False_ind;omega.
            reflexivity.
          }
        }
        {
          simpl.
          intros n1 H. 
          inversion_clear H. 
          inversion_clear H0.
          case (eq_nat_dec n n0).
          { intros abs;elim H1;subst; reflexivity. }
          { reflexivity. }
        }
      *  apply And_e_r with (Var n0);apply Ax;simpl;auto.

 - intro Hdeduce.
   case (eq_nat_dec n0 n).
   + intro;subst.
     apply And_e_l with (build_from_word_from (S n) l0).
     apply Ax;simpl;auto.
   + intro H1.
     apply Impl_e with (build_from_word_from (S n0) l0).
     * apply Aff with nil;[red;simpl;auto|].
       apply Impl_i.
       apply IHf.
       intros I H_I.
       red.
       rewrite change_interp with (I':=fun x => if eq_nat_dec x n0 then false else I x).       
       { apply Hdeduce.
         red;simpl;intros F [h|h];[subst|tauto].
         red;simpl.
         case (eq_nat_dec n0 n0);[|intro abs;elim abs;reflexivity].
         intros _;simpl.
         rewrite change_interp with (I':=I).
         { apply H_I.
           simpl;auto.
         }
         { intros n1 H0.
           apply build_from_word_from_vars in H0.
           destruct H0.
           case (eq_nat_dec n1 n0).
           intro;apply False_ind;omega.
           reflexivity.
         }
       }
       {
         intros n1 H;inversion_clear H.
         inversion_clear H0.
         case (eq_nat_dec n n0).
         { intros abs;elim H1;subst;reflexivity. }
         { reflexivity. }
       }
     *  apply And_e_r with (¬ (Var n0)).
        apply Ax;simpl;auto.
Qed.


(* Only one interpretation (up to the variables appearing in Φʷ) is satisfying 
   Φʷ
*)
Lemma build_from_word_from_only_deduced_from_interp_from_word_from n w : 
  forall I, satisfy_formula I (build_from_word_from n w) -> 
            forall x, appears_in_formula x (build_from_word_from n w) -> 
                      I x = interp_from_word_from n w x .
Proof.

  functional induction (build_from_word_from n w);
  intros I H x Hin.

  - inversion Hin. 
    inversion H1.

  - red in H; simpl in *.
    rewrite Bool.andb_true_iff in H.
    destruct H as [H1 H2].
    inversion_clear Hin.
    + inversion_clear H. 
      subst.
      case (eq_nat_dec n n);[|intro abs;elim abs;reflexivity].
      auto.
    + case (eq_nat_dec n x).
      * intro abs.
        apply build_from_word_from_vars in H.
        apply False_ind;omega.
      * auto.

  - red in H; simpl in *.
    rewrite Bool.andb_true_iff in H.
    destruct H as [H1 H2].
    inversion_clear Hin.
    + inversion_clear H. 
      inversion_clear H0.
      subst.
      case (eq_nat_dec n n);[|intro abs;elim abs;reflexivity].
      rewrite Bool.negb_true_iff in H1;auto.
    
    + case (eq_nat_dec n x).
      * intro abs.
        apply build_from_word_from_vars in H.
        apply False_ind;omega.
      * auto.
Qed.

(* Thus any formula which variables are included in build_from_word_from 
   has a unique value for all formula satisfing build_from_word_from 
*)
Lemma build_from_word_from_constains_formula n w : 
  forall I F, incl (var_set F) (var_set (build_from_word_from n w)) -> 
              satisfy_formula I (build_from_word_from n w) -> 
              interp I F = interp (interp_from_word_from n w) F.
Proof.

  intros I F H H1.
  apply change_interp.
  intros n0 H2.
  apply build_from_word_from_only_deduced_from_interp_from_word_from. 
  exact H1.
  rewrite <- var_set_equiv_appears_in_formula.
  apply H. 
  rewrite var_set_equiv_appears_in_formula.
  assumption.
Qed.

(* if one can deduce F₁∨F₂ from a build_from_word 
   then she can deduce either F₁ or F₂ from this build_from_word.
*)
Lemma deduce_destruct_or_build_word :
  forall w F1 F2
         (H_included: incl 
                        ((var_set F1)++(var_set F2))
                        (var_set (build_from_word w))), 
    deduce ((build_from_word w) :: nil) (F1 ∨ F2) ->
    (deduce (build_from_word w :: nil) F1)\/(deduce (build_from_word w :: nil) F2).
Proof.
  intros w F1 F2 H_included H.
  case_eq (interp (interp_from_word_from 0 w) F1);intro H1.
  {
    left.
    intros I H2.
    red;simpl.
    rewrite change_interp with (I':=interp_from_word_from 0 w).
    - exact H1.
    - intros n H3.
      apply build_from_word_from_only_deduced_from_interp_from_word_from. 
      + apply H2.
        auto with *.
      + rewrite <- var_set_equiv_appears_in_formula in H3|-*.
        apply H_included.
        auto with *.
  }
  {
    right.
    intros I H2.
    red;simpl.
    rewrite change_interp with (I':=interp_from_word_from 0 w).
    {
      assert (H3:=H (interp_from_word_from 0 w)). 
      unfold satisfy_formula in H3;simpl in H3.
      rewrite H1 in H3;simpl in H3.
      apply H3.
      red;simpl; intros F [h|h];[subst|tauto].
      apply interp_from_word_from_satisfies_build_from_word.
    }
    {
      intros n H3.
      apply build_from_word_from_only_deduced_from_interp_from_word_from.
      - apply H2;simpl;left;reflexivity. 
      - rewrite <- var_set_equiv_appears_in_formula in H3|-*.
        apply H_included.
        auto with *.
    }
  }
Qed.

(* We can now show that :                                                 *)
(*    a- if Φʷ ⊧ F then Φʷ ⊢ F                                            *)
(*    b- if Φʷ ⊧ ¬F then Φʷ ⊢ ¬F                                          *)
Lemma deduce_build_from_word_F_prove_build_from_word_F F :
  forall w (H_included: incl (var_set F) (var_set (build_from_word w))), 
    (deduce ((build_from_word w) :: nil) F ->
     ((build_from_word w :: nil)⊢ F)) /\
    (deduce ((build_from_word w) :: nil) (¬ F) ->
     ((build_from_word w :: nil)⊢ ¬F ))
.
Proof.  

  induction F as 
      [ | (* bot *)
        n | (* Var *)
        F1 IH1 F2 IH2 | (* AND *)
        F1 IH1 F2 IH2 | (* Or *)
        F1 IH1 F2 IH2 | (* impl *)
        F IH  (* NEG *)
      ];intros w  H_included.


  - split. (* bot *)
    + intros abs. (* absurd case *)
      red in abs.
      unfold satisfy_formula in abs;simpl in abs. 
      absurd (true=false).
      * discriminate.
      * symmetry;apply abs with (I:=(interp_from_word_from 0 w)).
        unfold satisfy_formula_set;simpl.
        intros F [H|abs'];[|tauto].
        subst;apply interp_from_word_from_satisfies_build_from_word. 

    + (* trivial case *)
      intros _.
      apply Neg_i.
      apply Ax;simpl;auto.

  - split. (* Var *) 
    + apply var_deduce_build_from_var_proved_build_from. 
    + apply neg_var_deduce_build_from_neg_var_proved_build_from. 

  - split. (* AND *)
    + intros H.
      destruct (deduce_destruct_and _ _ _ H) as [H1 H2].
      clear H.
      apply And_i.
      * destruct (IH1 w) as [H3 H4].
        {
          rewrite var_set_equation in H_included.
          destruct (incl_destruct_app _ _ _ _ H_included);assumption.
        }
        { 
          auto.
        }
      * destruct (IH2 w).
        {
          rewrite var_set_equation in H_included.
          destruct (incl_destruct_app _ _ _ _ H_included);assumption.
        }
        { auto. }
    + intros H.
      rewrite deduce_semantic_equivalence  with (F':= ¬ F1 ∨ ¬ F2) in H.
      * apply Neg_i. 
        simpl in H_included.
        destruct (deduce_destruct_or_build_word w (¬ F1) (¬ F2)  H_included H) as
            [H1 | H1].
        { 
            apply Neg_e with F1.
            { apply Aff with (build_from_word w :: nil);[auto with *|].
              destruct (IH1 w) as [H2 H3].
              { red. 
                auto with *. 
              }
              { auto. }
            }
            { apply And_e_l with F2;apply Ax;simpl;auto. }
          }
          { 
            apply Neg_e with F2.
            { apply Aff with (build_from_word w :: nil); [auto with *|]. 
              destruct (IH2 w) as [H2 H3].
              { red;auto with *.
              }
              { auto. }
            }
            { apply And_e_r with F1;apply Ax;simpl;auto. }
          }
      * apply equivalence_not_and_or_not.        

  - split. (* Or *)
    + intros H.
      destruct (deduce_destruct_or_build_word _ _ _ H_included H)
      as [H1 | H1].
      * apply Or_i_l.
        destruct (IH1 w) as [H2 H3].
       {intros x H2.
        apply H_included.
        simpl;auto with *. 
       }
       { auto. }
      * apply Or_i_r.
        destruct (IH2 w) as [H2 H3].
        { intros x H2.
          apply H_included.
          simpl;auto with *.
        }
        { auto. }
    + intros H.
      rewrite deduce_semantic_equivalence  with (F':= ¬ F1∧ ¬ F2) in H. 
      * destruct (deduce_destruct_and _ _ _ H) as [H1 H2].
        apply Neg_i.
        apply Or_e with F1 F2.
        { apply Ax;simpl;auto. 
        }
        { 
          apply Neg_e with F1.
          { apply Aff with (build_from_word w :: nil);[auto with *|].
            destruct (IH1 w) as [H3 H4].
            {
              intros x H3;apply H_included;simpl;auto with *.
            }
            {
              apply H4;auto.
            }
          }
          { apply Ax;simpl;auto.
          }
        }
        { apply Neg_e with F2.
          { apply Aff with (build_from_word w :: nil);[auto with *|].
            destruct (IH2 w) as [H3 H4].
            {
              intros x H3;apply H_included;simpl;auto with *.
            }
            {
              apply H4;auto.
            }
          }
          { apply Ax;simpl;auto.
          }
        }
      * apply equivalence_not_or_and_not.

  - split. (* FImpl *)
    + rewrite deduce_semantic_equivalence with (F':= ¬ F1 ∨ F2).
      { intros H.
        destruct (deduce_destruct_or_build_word _ (¬ F1) _  H_included H) 
          as [H1 | H1].
        { (* Neg F1 is true *)
          apply Impl_i;
          apply Abs_c.
          apply Neg_e with F1.
          { apply Aff with (build_from_word w :: nil);[auto with *|].
            destruct (IH1 w).
            {
              intros x H2;apply H_included;simpl;auto with *.
            }
            { auto.
            }
          }
          {
            apply Ax;simpl;auto.
          }
        } 
        { (* F2 is true *)
          apply Impl_i.
          apply Aff with (build_from_word w :: nil);[auto with *|].
          destruct (IH2 w).
            {
              intros x H2;apply H_included;simpl;auto with *.
            }
            { auto. }
        }
      }
      { apply equivalence_impl_or_not. }
    + rewrite deduce_semantic_equivalence with (F':=F1 ∧ (¬ F2)).
      { intros H.
        destruct (deduce_destruct_and  _ _ _ H) as [H1 H2].
        apply Neg_i.
        apply Neg_e with F2.
        { apply Aff with (build_from_word w :: nil);[auto with *|].
          destruct (IH2 w).
            { intros x H3;apply H_included;simpl;auto with *. }
            { auto. }
        }
        {
          apply Impl_e with F1.
          { apply Ax;simpl;auto. }
          { apply Aff with (build_from_word w :: nil);[auto with *|]. 
            destruct (IH1 w).
            { intros x H3;apply H_included;simpl;auto with *. }
            { auto. }
          }
        }
      }
      { apply equivalence_not_impl_and_not. }

  - split. (* Neg *)
    + intros H.
      destruct (IH w).
      { intros x H3;apply H_included;simpl;auto with *. }
      { auto. }
    + rewrite deduce_semantic_equivalence with (F':=F).
      {
        intros H.
        apply Neg_i.
        apply Neg_e with F.
        { apply Ax;simpl;auto. }
        {
          apply Aff with (build_from_word w :: nil);[auto with *|].
          destruct (IH w).
          { intros x H3;apply H_included;simpl;auto with *. }
          { auto. }
        }
      }
      { apply equivalence_neg_idempotent. }
Qed.


(* We can always find a word w such that 
   the variables of F are included in the variables of Φʷ
*)
Lemma exists_build_from_word F : 
  exists n, forall w, length w > n ->  
                      incl (var_set F) (var_set (build_from_word w)).
Proof.
  assert (exists n, forall x, In x (var_set F) -> x <= n).
  
  - induction F.
    + simpl. exists 0; tauto.
    + simpl.
      exists n.
      intros x [h|h];[subst;auto with *|tauto].
    + destruct IHF1 as [nF1 HF1];destruct IHF2 as [nF2 HF2].
      exists (max nF1 nF2).
      assert (h1:=Max.le_max_l nF1 nF2).
      assert (h2:=Max.le_max_r nF1 nF2).
      simpl.
      intros x H.
      apply in_app_or in H.
      destruct H.
      * apply HF1 in H.
        omega.
      * apply HF2 in H.
        omega.
    + destruct IHF1 as [nF1 HF1];destruct IHF2 as [nF2 HF2].
      exists (max nF1 nF2).
      assert (h1:=Max.le_max_l nF1 nF2).
      assert (h2:=Max.le_max_r nF1 nF2).
      simpl.
      intros x H.
      apply in_app_or in H.
      destruct H.
      * apply HF1 in H.
        omega.
      * apply HF2 in H.
        omega.
    + destruct IHF1 as [nF1 HF1];destruct IHF2 as [nF2 HF2].
      exists (max nF1 nF2).
      assert (h1:=Max.le_max_l nF1 nF2).
      assert (h2:=Max.le_max_r nF1 nF2).
      simpl.
      intros x H.
      apply in_app_or in H.
      destruct H.
      * apply HF1 in H.
        omega.
      * apply HF2 in H.
        omega.
    + destruct IHF as [nF HF].
      exists ( nF).
      simpl.
      intros x H.
      auto.

  -  destruct H as [n Hn].
     exists n.
     intros w H.
     intros x H1.
     rewrite var_set_equiv_appears_in_formula.
     unfold build_from_word.
     rewrite build_from_word_from_vars.
     simpl.
     apply Hn in H1.
     omega.
Qed.


Require Import Recdef.
(* build_from_word_from_vars m n construct the set of all words os size m-n  *)
Function build_all_word_from m n {measure (fun x => m - x) n} := 
  if le_lt_dec m n 
  then ((Var n∧⊤)::(¬ (Var n)∧⊤)::nil)
  else (List.map (fun (F:Formula) => ( (Var n) ∧ F)) (build_all_word_from m (S n)))
         ++
         (List.map (fun (F:Formula) => ((¬ (Var n)) ∧ F)) (build_all_word_from m (S n)))
.
Proof.
  intros m n anonymous teq.
  omega.
Qed.

Definition build_all_word m := build_all_word_from m 0.


(* Link between build_all_word_from and build_from_word_from *)
Lemma build_all_word_from_correct :
  forall m n F, In F (build_all_word_from m n) -> 
                exists w, length w = S(m-n)/\ F=build_from_word_from n w.
Proof.
  intros m n.
  functional induction (build_all_word_from m n).

  - intros F H. simpl in H;destruct H as [H|[H|H]].
    + subst; exists (true::nil).
      split.
      * simpl. omega.
      * simpl. reflexivity.
    + subst; exists (false::nil).
      split.
      * simpl. omega.
      * simpl;reflexivity.
    + tauto.

  - intros F H1.
    apply in_app_or in H1.
    destruct H1 as [H1|H1].
    + rewrite in_map_iff in H1.
      destruct H1 as [x [H1 H2]].
      subst.
      destruct (IHl _ H2) as [w [H3 H4]].
      exists (true::w).
      split.
      * simpl;omega.
      * subst. simpl. destruct w.
        { apply False_ind;simpl in H3;omega. }
        { reflexivity. }
    + rewrite in_map_iff in H1.
      destruct H1 as [x [H1 H2]].
      subst.
      destruct (IHl _ H2) as [w [H3 H4]].
      exists (false::w).
      split.
      * simpl;omega.
      * subst. simpl. destruct w.
        { apply False_ind;simpl in H3;omega. }
        { reflexivity. }
Qed.

(* We can prove that ⊢ ⋁Φʷ *)
Lemma provable_build_all_word :
  forall n m, 
    ( nil ⊢ build_big_or (build_all_word_from m n)).
Proof.
  intros n m.
  functional induction (build_all_word_from m n).

  - simpl.
    apply excluded_middle with (Var n).
    + apply Or_i_l.
      apply And_i.
      * apply Ax;simpl;auto.
      * apply Proof_true.
    + apply Or_i_r.
      apply Or_i_l.
      apply And_i.
      * apply Ax;simpl;auto.
      * apply Proof_true.

  - rewrite Proof_build_big_or_app.
    apply excluded_middle with (Var n).
    + apply Or_i_l.
      rewrite Proof_Or_andl_distrib.
      apply And_i.
      * apply Ax;simpl;auto.
      * apply Aff with nil;auto with *.
    + apply Or_i_r.
      rewrite Proof_Or_andl_distrib.
      apply And_i.
      * apply Ax;simpl;auto.
      * apply Aff with nil;auto with *.
Qed.

(* Last effort *)
Lemma tautologies_completeness F : 
  tautology F -> (nil ⊢ F).
Proof.
  intros H.
  destruct (exists_build_from_word F) as [n H1].
  apply Extended_Or_e with (l:=build_all_word (S n)).

  - unfold build_all_word. 
    apply provable_build_all_word.

  - intros F' H0.
    apply build_all_word_from_correct in H0.
    simpl in H0.
    destruct H0 as [ w [H2 H3]].
    subst.
    apply deduce_build_from_word_F_prove_build_from_word_F.
    + apply H1.
      omega.
    + intros I _.
      apply H.
Qed.

(* Now we can finish the proof *)
Lemma completeness Γ F :
  deduce Γ F -> (Γ ⊢ F).
Proof.
  revert F.
  induction Γ;intro F.
  
  - intros H.
    apply tautologies_completeness.
    intros I.
    apply H.
    red;simpl.
    tauto.

  - intros H.
    apply Impl_e with a.
    + apply Aff with Γ;[auto with *|]. 
      apply IHΓ.
      apply deduce_impl_intro. assumption.
    + apply Ax;simpl;auto.
Qed.
