(**************************************************************************)
(*                                                                        *)
(*     A tutorial on propositional logic and its properties               *)
(*     Copyright (C) 2013                                                 *)
(*                                                                        *)
(*                                                                        *)
(*     Julien Forest                                                      *)
(*                                                                        *)
(*     Cedric-ENSIIE                                                      *)
(*                                                                        *)
(*                                                                        *)
(*   This file is distributed under the terms of the CeCILL-C licence     *)
(*                                                                        *)
(**************************************************************************)

(** 
    This module defines the semantic of propositional logic 
*)

Require Import Formula.
Require Import Bool.
Require Import List.
Require Import Arith.
Require Import Utf8_core.
(* Variables interpretation are defined as function from 
   nat to bool
 *)
Definition var_interp := nat -> bool.

(* We can extend an interpretation from nat to formula *)
Function interp (I: var_interp) (F:Formula) := 
  match F with 
      | Bot => false
      | Var p => I p
      | F ∧ G => (interp I F) && (interp I G)
      | F ∨ G => (interp I F) || (interp I G)
      | F → G => implb (interp I F) (interp I G)
      | ¬ F => negb (interp I F)
  end.

(* We then define basic semantic notions *)
Definition satisfy_formula I F := 
  (* an interpretation I satisfies a formula F iff *)     
  interp I F = true.

Definition satisfy_formula_set I Γ := 
  (* an interpretation I satifies a formula set Γ iff 
     it satifies all then formulas in Γ*)
  forall F,  In F Γ -> satisfy_formula I F.

(* a tautology is satisfied by all interpretation *)
Definition tautology F := forall I, satisfy_formula I F.

(* A formula set Γ is contradictory iff 
   forall interpretation we can find a formula in Γ 
   which is not satisfied by F
*)
Definition contradictory Γ := 
  forall I, exists F, In F Γ /\ ~ satisfy_formula I F.

(* A formula F can be deduce from a formula set Γ iff
   any interpretation satisfying Γ satisfies F *)
Definition deduce Γ F := 
  forall I, satisfy_formula_set I Γ -> satisfy_formula I F.

(* Equivalence as presented in course *)
Definition semantic_equivalence F G := 
  deduce (F::nil) G /\ deduce (G::nil) F.

Notation " F ≡ G " := (semantic_equivalence F G) (at level 100,no associativity) : Logic_scope.

(* Another definition of Equivalence *)
Definition semantic_equivalence_2 F G := forall I, interp I F = interp I G.

(* Both definition of equivalence are equivalent *)
Lemma semantic_equivalence_equiv : 
  forall F G, semantic_equivalence F G <-> semantic_equivalence_2 F G.
Proof.
  intros F G.
  split.
  - intros [H1 H2] I.
    red in H1,H2.
    case_eq (interp I F).
    + intros H3.
      symmetry.
      apply H1.
      red.
      intros F' [H4 | H4].
      subst;exact H3.
      simpl in H4; tauto.
    + intros H3.
      case_eq (interp I G).
      intro H4.
      * replace (interp I F) with true in H3.
        symmetry;assumption.
        symmetry.
        apply H2.
        intros F' [H5 | H5].
        subst;exact H4.
        simpl in H5; tauto.
      *  tauto.      

  - intros H;red in H.
    split.
    +  intros I H1.
       red. rewrite <- H.
       apply H1;simpl;auto.
    +  intros I H1.
       red. rewrite  H.
       apply H1;simpl;auto.
Qed.


(* We can deduce G from  {F→G, F} *)
Lemma deduce_impl :
  forall F G, deduce (F::(F→G)::nil) G.
Proof.
  intros F G I H.  
  red in H;simpl in H.
  assert (I_sat_FG:satisfy_formula I (F→ G)) by auto.
  red in I_sat_FG;simpl in I_sat_FG.
  assert (I_sat_F:satisfy_formula I F) by auto.
  red in I_sat_F;simpl in I_sat_F.
  rewrite I_sat_F in I_sat_FG.
  exact I_sat_FG.
Qed.

(* We can check if a formula set is satisfied by a given 
   interpretation 
*)
Lemma satisfy_formula_set_dec : 
  forall I Γ, 
    {satisfy_formula_set I Γ}+
    {~ satisfy_formula_set I Γ}.
Proof.
  intros I Γ.
  induction Γ as [| F Γ IH].

  - left;red;simpl;tauto.
  - (* we do a case analysis on the satisfiability of Γ *)
    destruct IH as [H | H].
    + (* if Γ is satisfied by I 
         we test if F is satisfied or not
       *)
      case_eq (interp I F); intros H1.
      * left.
        red.
        intros F' H2.
        simpl in H2. 
        destruct H2 as [ H2 | H2].
        { subst. 
          apply H1.
        }
        { apply H. 
          assumption.
        }
      * right. 
        intros abs.
        red in abs.
        rewrite (abs F) in H1.
        discriminate H1.
        simpl;auto.
    + right. 
      intros abs.
      red in abs.
      apply H.
      intros F' HF'.
      apply abs.
      simpl;auto.
Qed.

(* If I does not satisfies Γ then we can find a formula 
   F in Γ not satisfied by I
*)
Lemma destruct_not_satisfied I Γ : 
  ~ satisfy_formula_set I Γ -> 
  exists F, In F Γ /\ ~satisfy_formula I F.
Proof.
  induction Γ as [ | F Gamma IH].

  - intro abs.
    elim abs.
    intros F.
    simpl;tauto.

  - intros H.
    case_eq (interp I F);intros H1.
    + destruct IH as [ G [ H2 H3]].
      * intros H2;elim H.
        intros G H3.
        simpl in H3.
        destruct H3 as [H3|H3].
        {subst.
         exact H1.
        }
        {apply H2.
         assumption.
        }
      * exists G.
        split.
        { simpl; auto. }
        { assumption. }
    + exists F.
      split.
      simpl;auto.
      intro H2;rewrite H2 in H1;discriminate.
Qed.

Lemma deduce_impl_intro Γ F G : 
  deduce Γ (F → G) <-> deduce (F::Γ) G.
Proof.
  split.
  - (* left to right case *)
    intros H I H1.
    assert (H2:satisfy_formula_set I Γ) by 
        (intros F1 H_F1; apply H1;simpl;auto).
    apply deduce_impl with F.
    intros F' H3.
    simpl in H3.
    destruct H3 as [H3 | [ H3 | H3]].
    + apply H1; simpl;auto.
    + subst; auto.
    + tauto.

  - (* right to left case *)
    intros H I H1.
    red;simpl.
    case_eq (interp I F);intros H2.
    +  simpl. 
       apply H.
       red;simpl. intros F' [H3|H3].
       subst;exact H2.
       apply H1;assumption.
    + simpl;reflexivity.
Qed.

Lemma deduce_is_contradictory_neg Γ F :
  deduce Γ F <-> contradictory ((¬ F)::Γ).
Proof.
  split.

  - (* left to right case *)
    intros H I.
    case (satisfy_formula_set_dec I Γ);intros H1.
    + (*  if Γ is satisfied by I *)
      exists (¬ F).
      split.
      * simpl;auto.
      * intros H2.
        red in H2.
        simpl in H2.
        rewrite negb_true_iff in H2.
        assert (H3:=H I H1).
        rewrite H3 in H2. 
        discriminate.
    + (* if Γ is not satisfied by I*)
      destruct (destruct_not_satisfied _ _ H1) as [F' [H2 H3]].
      exists F';split.
      * simpl;auto.
      * assumption.

  - (* right to left case *)
    intros H I H1.
    apply not_false_is_true.
    intros H2.
    red in H.
    destruct (H I) as [F' [H3 H4]].
    simpl in H3.
    destruct H3 as [H3 | H3].
    + subst. apply H4.
      red. simpl. rewrite H2. reflexivity.
    + apply H4.
      apply H1.
      exact H3.
Qed.

(* we can define a new interpretation from an interpretation I by 
   setting it value on variable p
*)
Definition change_value (I:var_interp) p v := 
  fun q => if eq_nat_dec p q then v else I q.


(* The next lemma explain how to compute the 
   value of a replacement from the value 
   of the original formula 
*)
Lemma replace_interp I F p G : 
  interp I (freplace F p G) = interp (change_value I p (interp I G)) F.
Proof.
  induction F as 
      [ | q | F1 IH1 F2 IH2 | F1 IH1 F2 IH2 | F1 IH1 F2 IH2 | F IH ].

  - reflexivity.

  - simpl.
    unfold change_value.
    case (eq_nat_dec p q).
    + reflexivity.
    + reflexivity.

  - simpl. 
    rewrite IH1;rewrite IH2. reflexivity.

  - simpl. rewrite IH1;rewrite IH2. reflexivity.

  - simpl. rewrite IH1;rewrite IH2. reflexivity.

  - simpl. rewrite IH. reflexivity.
Qed.  

(* tautologies are preserved by replacement *)
Lemma replace_tautology F p G: 
  tautology F -> tautology (freplace F p G).
Proof.
  intros HF I.
  red. rewrite replace_interp.
  apply HF.
Qed.


(* interpretation if stable by equivalence on both sides*)
Lemma replace_equiv_l I F F' p G : 
  semantic_equivalence F F' -> 
  interp I (freplace F p G) = interp I (freplace F' p G).
Proof.
  intros Heq.

  do 2 rewrite replace_interp.
  rewrite semantic_equivalence_equiv in Heq.
  apply Heq.
Qed.

Lemma replace_equiv_r I F p G G' : 
  semantic_equivalence G G' -> 
  interp I (freplace F p G) = interp I (freplace F p G').
Proof.
  intros Heq.

  do 2 rewrite replace_interp.
  rewrite semantic_equivalence_equiv in Heq.
  rewrite Heq;reflexivity.
Qed.

(* Γ deduces ⊥  iff it is contradictory *)
Lemma deduce_bot_contradictory Γ : 
  deduce Γ Bot <-> contradictory Γ.
Proof.

  rewrite deduce_is_contradictory_neg .
  split.

  - intros H I.
    destruct (H I) as [F [H1 H2]].
    simpl in H1;destruct H1 as[H1|H1].
    + subst.
      elim H2.
      reflexivity.
    + exists F; auto.
  
  - intros H I.
    destruct (H I) as [F [H1 H2]].
    exists F;simpl;auto.
Qed.
 
(* deduce is stable by equivalence *)
Lemma deduce_semantic_equivalence_2 Γ F F' :
  semantic_equivalence_2 F F' ->
  (deduce Γ F <-> deduce Γ F').
Proof.
  intro H.
  split.

  - intros H1 I H_I.
    red;rewrite <- H.
    apply H1;assumption.

  - intros H1 I H_I.
    red;rewrite H.
    apply H1;assumption.
Qed.

Lemma deduce_semantic_equivalence Γ F F' :
  semantic_equivalence F F' -> 
  (deduce Γ F <-> deduce Γ F').
Proof.
  intros H.
  rewrite semantic_equivalence_equiv in H.
  apply deduce_semantic_equivalence_2;assumption.
Qed.


(* We can freely change interpretation values on variables 
   which does not belong to a formula F without changing its
   interpretation
*)
Lemma change_interp  I I' F : 
  forall (I_I':forall n, appears_in_formula n F -> I n = I' n),
  interp I F = interp I' F.
Proof.
  induction F as 
      [ | (* bot *)
        n | (* Var *)
        F1 IH1 F2 IH2 | (* AND *)
        F1 IH1 F2 IH2 | (* Or *)
        F1 IH1 F2 IH2 | (* impl *)
        F IH  (* NEG *)
      ];intro H.

  - reflexivity.

  - apply H. apply Formula_var.

  - simpl. rewrite IH1;[rewrite IH2|].
    + reflexivity.
    + intros n H1.
      apply H.
      apply Formula_and_r.
      assumption.
    + intros n H1.
      apply H.
      apply Formula_and_l.
      assumption.

  - simpl;rewrite IH1;[rewrite IH2|].
    + reflexivity.
    + intros n H1.
      apply H.
      apply Formula_or_r.
      assumption.
    + intros n H1.
      apply H.
      apply Formula_or_l.
      assumption.

  - simpl;rewrite IH1;[rewrite IH2|].
    + reflexivity.
    + intros n H1.
      apply H.
      apply Formula_impl_r.
      assumption.
    + intros n H1.
      apply H.
      apply Formula_impl_l.
      assumption.

  - simpl;rewrite IH.
    + reflexivity.
    + intros n H1.
      apply H.
      apply Formula_neg.
      assumption.
Qed.

(* if Γ deduces F1∧F2 then Γ deduces F1 and Γ deduces F2 *)
Lemma deduce_destruct_and Γ F1 F2 : 
  deduce Γ (FAnd F1 F2) -> (deduce Γ F1)/\(deduce Γ F2).
Proof.
  intros H.
  split.

  -  intros I H'.
     apply H in H'.
     red in H';simpl in H'.
     apply andb_true_iff in H'.
     destruct H'.
     apply H0.

  -  intros I H'.
     apply H in H'.
     red in H';simpl in H'.
     apply andb_true_iff in H'.
     destruct H'.
     apply H1.
Qed.


(* Some classical equivalences *)

Lemma equivalence_and_idempotent F : F∧F ≡ F.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);reflexivity.
Qed.

Lemma equivalence_or_idempotent F : F∨F ≡ F.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);reflexivity.
Qed.

Lemma equivalence_and_commute F G : F∧G ≡ G∧F.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    destruct (interp I G);
    reflexivity.
Qed.

Lemma equivalence_or_commute F G : F∨G ≡ G∨F.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    destruct (interp I G);
    reflexivity.
Qed.

Lemma equivalence_and_assoc F G H : F∧(G∧H) ≡ (F∧G)∧H.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    destruct (interp I G);
    destruct (interp I H);
    reflexivity.
Qed.

Lemma equivalence_or_assoc F G H : F∨(G∨H) ≡ (F∨G)∨H.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    destruct (interp I G);
    destruct (interp I H);
    reflexivity.
Qed.

Lemma equivalence_and_or_distrib F G H : F∧(G∨H) ≡ (F∧G)∨(F∧H).
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    destruct (interp I G);
    destruct (interp I H);
    reflexivity.
Qed.


Lemma equivalence_or_and_distrib F G H : F∨(G∧H) ≡ (F∨G)∧(F∨H).
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    destruct (interp I G);
    destruct (interp I H);
    reflexivity.
Qed.

Lemma equivalence_not_and_or_not F G : ¬ (F∧G) ≡ ¬ F ∨ ¬G.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    destruct (interp I G);
    reflexivity.
Qed.

Lemma equivalence_not_or_and_not F G : ¬ (F∨G) ≡ ¬ F ∧ ¬G.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    destruct (interp I G);
    reflexivity.
Qed.

Lemma equivalence_bot_and_bot F : (F∧⊥) ≡ ⊥.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    reflexivity.
Qed.

Lemma equivalence_bot_or_F F : (F∨⊥) ≡ F.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    reflexivity.
Qed.

Lemma equivalence_top_or_top F : (F∨⊤) ≡ ⊤.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    reflexivity.
Qed.

Lemma equivalence_top_and_F F : (F∧⊤) ≡ F.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    reflexivity.
Qed.

Lemma equivalence_impl_or_not F G : (F→G) ≡ ¬ F ∨ G.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    destruct (interp I G);
    reflexivity.
Qed.


Lemma equivalence_not_impl_and_not F G : ¬ (F→G) ≡ F ∧ ¬ G.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    destruct (interp I G);
    reflexivity.
Qed.

Lemma equivalence_neg_idempotent F : ¬ ¬ F ≡ F.
Proof.
  rewrite semantic_equivalence_equiv.
  intros I.
  simpl. 
  destruct (interp I F);
    reflexivity.
Qed.
